/*
 * debug.c - Debugger interface
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "dwarfx.h"
#include "ptrace.h"

#include "util.h"
#include "error.h"
#include "data.h"
#include "var.h"
#include "op.h"
#include "stmt.h"
#include "sbug.h"


extern struct dwx_obj *objects;

static struct pt_process *proc;


struct object {
    struct pt_process *process;
    const struct dwx_obj *obj;
    unsigned long addr;
};


/* ----- Values ------------------------------------------------------------ */


struct dwx_type proc_type;
struct dwx_type object_type;


struct value *proc_value(struct pt_process *proc)
{
    struct value *value;

    value = new_value(&proc_type);
    value->value = proc;
    return value;
}


struct pt_process *value_proc(struct value *value)
{
    struct pt_process *res;

    if (!value || value->type != &proc_type)
        clerrorf("invalid type (expected process)");
    res = (struct pt_process *) value->value;
    put(value);
    return res;
}


struct pt_process *proc_eval(const struct op *op)
{
    return value_proc(evaluate(op));
}


struct value *object_value(struct object *obj)
{
    struct value *value;

    value = new_value(&object_type);
    value->value = obj;
    return value;
}


const struct object *value_object(struct value *value)
{
    const struct object *res;

    if (!value || value->type != &object_type)
        clerrorf("invalid type (expected object)");
    res = value->value;
    put(value);
    return res;
}


const struct object *object_eval(const struct op *op)
{
    return value_object(evaluate(op));
}


/* ----- Operators --------------------------------------------------------- */


/*
 * Design issue: is this the right moment for binding an object to an address
 * space (process) ? Or should this be deferred to the time of evaluation ?
 * Normally, there isn't much of a difference, of course.
 */


struct value *op_object(const struct op *self)
{
    const char *path = self->u.arg.a ? string_eval(self->u.arg.a) : NULL;
    const char *name = self->u.arg.b ? string_eval(self->u.arg.b) : NULL;
    const char *object_name = string_eval(self->u.arg.c);
    const struct dwx_obj *obj;
    struct object *object;

    obj = objects;
    if (path) {
	obj = dwx_find_unit(objects,path);
	if (!obj) clerrorf("no such unit \"%s\"",path);
    }
    if (name) {
	obj = dwx_find_item(obj->children,name);
	if (!obj) clerrorf("no such parent object \"%s\"",name);
    }
    obj = dwx_find_item(obj->children,object_name);
    if (!obj) clerrorf("no such object \"%s\"",object_name);
    object = alloc_t(struct object);
    object->process = proc;
    object->obj = obj;
    object->addr = obj->low_pc;
    return object_value(object);
}


/* ----- Statements -------------------------------------------------------- */


void stmt_continue(const struct stmt *self)
{
    struct pt_process *proc;

    proc = proc_eval(self->u.breakpoint.process);
    pt_run(proc);
    /*
     * @@@ Not very flexible ...
     */
    (void) pt_wait(1);
}


static unsigned long addr_eval(const struct op *op)
{
    struct value *value;
    const struct object *object;

    value = evaluate_lvalue(op);
    if (!value || value->type != &object_type) return value_int(value);
    object = value_object(value);
    if (object->obj->kind != dwx_obj_fn && object->obj->kind != dwx_obj_label)
	clerrorf("invalid object type (expected function or label)");
    return object->obj->low_pc;
}


void stmt_breakpoint(const struct stmt *self)
{
    struct pt_process *proc;
    unsigned long addr;

    proc = proc_eval(self->u.breakpoint.process);
    addr = addr_eval(self->u.breakpoint.location);
    pt_break(proc,addr);
}


void stmt_delete(const struct stmt *self)
{
    struct pt_process *proc;
    unsigned long addr;

    proc = proc_eval(self->u.breakpoint.process);
    addr = addr_eval(self->u.breakpoint.location);
    pt_unbreak(proc,addr);
}


int destroy_sbug_stmt(const struct stmt *stmt)
{
    if (stmt->fn == stmt_continue) {
	destroy_expr(stmt->u.breakpoint.process);
	return 1;
    }
    if (stmt->fn == stmt_breakpoint || stmt->fn == stmt_delete) {
	destroy_expr(stmt->u.breakpoint.process);
	destroy_expr(stmt->u.breakpoint.location);
	return 1;
    }
    else return 0;
}


/* ----- Data access ------------------------------------------------------- */


struct value *dereference_object(struct value *value)
{
    const struct object *object;
    const struct dwx_type *type;
    int size;
    struct value *new;
    void *data;

    object = value_object(value);
    if (object->obj->kind != dwx_obj_var)
	clerrorf("invalid object type (expected variable or constant)");
    type = object->obj->u.var.type;
    size = dwx_sizeof(type);
    new = new_value(type);
    data = alloc(size);
    /*
     * @@@ Optimize: create global owner of constant data, then use this as
     * new->owner
     */
    if (object->obj->u.var.constant)
	memcpy(data,object->obj->u.var.constant,size);
    else pt_read(object->process,
	  dwx_loc_here(object->obj->u.var.loc,object->process->regs.eip),
	  data,size);
		/* ia32-ism @@@ */
    new->value = data;
    return new;
}


/* ----- Helper functions for sbug.c --------------------------------------- */


static void kill_subprocess(void)
{
    pt_terminate(proc);
}


void launch_subprocess(char **argv)
{
    proc = pt_fork();
    if (proc) {
        struct var *var;

	atexit(kill_subprocess);
        var = var_lookup("proc");
        put(var->value);
        var->value = proc_value(proc);
    }
    else {
	if (execvp(*argv,argv) < 0) {
	    perror("execvp");
	    exit(1);
	}
	/* can't happen */
	exit(0);
    }
}
